/*	This is a simple module that takes in a 4 bit value and outputs
 *	one of 16 possible values. Though they are called 'frequency'
 *	they are in fact a period, in clock cycles, of the sine wave
 *	that will create a musical note
 */
module FrequencyDecoder(notes, frequency);
	input [3:0] notes;
	output reg [25:0] frequency;

	always@(*)
		case(notes)
			0: frequency = 26'd454544;
			1: frequency = 26'd429032;
			2: frequency = 26'd404952;
			3: frequency = 26'd382226;
			4: frequency = 26'd360772;
			5: frequency = 26'd320524;
			6: frequency = 26'd321410;
			7: frequency = 26'd303372;
			8: frequency = 26'd286346;
			9: frequency = 26'd270274;
			10: frequency = 26'd255104;
			11:	frequency = 26'd240786;
			12:	frequency = 26'd227272;
			13:	frequency = 26'd214516;
			14:	frequency = 26'd202476;
			15:	frequency = 26'd191114;
			default: frequency = 26'd270274;
		endcase
endmodule

/*	This module creates an approximation of a sinewave
 *	using 32 sampled points along the wave. These points are
 *	output at a rate dictated by the frequency (actually a period)
 *	in order to create a PCM (Pulse-Code Modulated) sine wave
 *	WARNING: because of the way this is done, the sampling rate
 *	varies based on frequncy, so simple wave addition to add
 *	these waves will not work correctly.
 */
module pcm_generator(clock, frequency, wave);
	input clock;
	input [26:0] frequency;
	output reg signed [20:0] wave;
//The 'frequency' passed to this module should actually be count of the period length of the wave in cycles
//the maximum value that can be reached is 1048575 (20-bit number) minimum is 0
//The 'zero' point should be half that (524287)

	wire signed [19:0] sine [31:0];
	assign sine[0] = 0;
	assign sine[1] = 102283;
	assign sine[2] = 200635;
	assign sine[3] = 291278;
	assign sine[4] = 370726;
	assign sine[5] = 435928;
	assign sine[6] = 484378;
	assign sine[7] = 514213;
	assign sine[8] = 514213;
	assign sine[9] = 484378;
	assign sine[10] = 435928;
	assign sine[11] = 370726;
	assign sine[12] = 291278;
	assign sine[13] = 200635;
	assign sine[14] = 102283;
	assign sine[15] = 0;
	assign sine[16] = 0;
	assign sine[17] = -102283;
	assign sine[18] = -200635;
	assign sine[19] = -291278;
	assign sine[20] = -370726;
	assign sine[21] = -435928;
	assign sine[22] = -484378;
	assign sine[23] = -514213;
	assign sine[24] = -514213;
	assign sine[25] = -484378;
	assign sine[26] = -435928;
	assign sine[27] = -370726;
	assign sine[28] = -291278;
	assign sine[29] = -200635;
	assign sine[30] = -102283;
	assign sine[31] = 0;

	reg [26:0] counter;
	reg [4:0] sin_ctr;

	//Every 32nd of the 'frequency' value, update the sine wave sampled point
	always@(posedge clock)
	begin
		if(counter >= (frequency >> 5)) //divide by 32
		begin
			counter <= 0;
			wave <= sine[sin_ctr];
			sin_ctr <= sin_ctr + 1;
		end
		else
		begin
			counter <= counter + 1;
			wave <= wave;
		end
	end
endmodule

/*	This module takes in a frequency and varies it up and down
 *	over time.
 */
module vibratoGenerator(clock, inFreq, outFreq, activate);
	input wire clock, activate;
	input wire [25:0] inFreq;
	output reg [25:0] outFreq;
	
	reg [25:0] counter;
	parameter vib_rate = 26'd20000000;

	always@(posedge clock)
	begin
		if(activate)
		begin
			if(counter >= vib_rate)
			begin
				outFreq <= inFreq;
				counter <= 0;
			end
			else
			counter <= counter +1;

			if(counter == 5000000)
				outFreq <= inFreq + (3500);
			if(counter == 10000000)
				outFreq <= inFreq;
			if(counter == 15000000)
				outFreq <= inFreq - (3500);
		end
		else
			outFreq <= inFreq;
	end
endmodule

/*	A simple module that multiples the input 'inFreq'
 *	by 2 when activate is high - since 'inFreq' is actually
 *	a period, doing this drops the produced tone by one octave.
 */
module octaveShift(clock, inFreq, outFreq, activate);
	input wire clock, activate;
	input wire [25:0] inFreq;
	output reg [25:0] outFreq;
	
	always@(posedge clock)
	begin
		if(activate)
			outFreq <= inFreq << 1;
		else
			outFreq <= inFreq;	
	end
endmodule
